﻿/*---------------------------------------------------------------------------*
  Project:  Horizon

  Copyright (C)Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 58466 $
 *---------------------------------------------------------------------------*/

#include "applet.h"

namespace
{
    void RestoreGraphicSetting()
    {
        // Recover the GPU register settings
        // However, the following is the method for recovering the state set using DMPGL API.
        // If using other libraries, it is necessary to reissue all of the register setting commands with their own methods.
        nngxUpdateState(NN_GX_STATE_ALL);

        // (Reference) Recovering settings with the GD library
        // nn::gd::System::ForceDirty(nn::gd::MODULE_ALL);

        // (Reference) Regarding restoring settings made using the GR library, the GR API is only for generating commands. Thus, it has no features for restoring settings. 
        // 
        // When rendering next time, run all of the necessary commands.
    }
}

bool TransitionHandler::s_IsExitRequired = false;
volatile bool TransitionHandler::s_IsAfterWakeUp = false;
nn::os::LightEvent TransitionHandler::s_AwakeEvent;
nn::os::CriticalSection TransitionHandler::s_CriticalSection;
nn::os::CriticalSection TransitionHandler::s_CriticalSectionForSleep;

TransitionHandler::TRANSITION_CALLBACK TransitionHandler::s_PrepareSleepCallback;
TransitionHandler::TRANSITION_CALLBACK TransitionHandler::s_AfterSleepCallback;
TransitionHandler::TRANSITION_CALLBACK TransitionHandler::s_PrepareHomeButtonCallback;
TransitionHandler::TRANSITION_CALLBACK TransitionHandler::s_AfterHomeButtonCallback;
TransitionHandler::TRANSITION_CALLBACK TransitionHandler::s_PreparePowerButtonCallback;

void TransitionHandler::Initialize()
{
    // Set sleep-related callbacks
    nn::applet::SetSleepQueryCallback(SleepQueryCallback, 0);
    nn::applet::SetAwakeCallback(AwakeCallback, 0);
//     nn::applet::SetSleepCanceledCallback(NULL, 0); // Recommend not using SleepCanceledCallback

    s_AwakeEvent.Initialize(true);
    s_CriticalSection.Initialize();
    s_CriticalSectionForSleep.Initialize();

    // During wake-up, always set to Signal state
    s_AwakeEvent.Signal();

    // Enable features related to the applet library.
    // If the parameter is set to false, sleep is automatically refused until nn::applet::EnableSleep() is called.
    // 
    nn::applet::Enable(false);

    // The Initialize function for each of the gx, snd, and dsp libraries must be called after the applet::Enable function
    // In particular, the nngxInitialize function, which initializes the GX library, should be called after determining whether to exit immediately following a call to the Enable function.

    // The application terminates here if the termination conditions are already established, such as by pressing the POWER Button while displaying the logo
    if ( nn::applet::IsExpectedToCloseApplication() )
    {
        s_IsExitRequired = true;
    }
}

void TransitionHandler::Finalize()
{
    s_AwakeEvent.Finalize();
    s_CriticalSection.Finalize();
    s_CriticalSectionForSleep.Finalize();

    nn::applet::SetSleepQueryCallback(NULL, 0);
    nn::applet::SetAwakeCallback(NULL, 0);
}

void TransitionHandler::EnableSleep()
{
    // Enable responses to sleep requests.
    // Check the system state. If the system has been closed, issue a sleep request.
    nn::applet::EnableSleep(true);
}

void TransitionHandler::DisableSleep()
{
    // Rejects the sleep mode. 
    // If a sleep request has already been sent, it returns a rejection. 
    nn::applet::DisableSleep(true);
}

void TransitionHandler::Process()
{
    // When waking up from sleep, enable LCD after waiting for one-frame worth of rendering to complete
    if ( TransitionHandler::s_IsAfterWakeUp )
    {
        nn::gx::StartLcdDisplay();
        TransitionHandler::s_IsAfterWakeUp = false;
    }

    // When considering the consistency of GPU register settings, responses to all types of transition and sleep requests should be done after rendering.
    // 

    // Determination for System Sleep Mode
    if ( nn::applet::IsExpectedToReplySleepQuery() )
    {
        // When REJECT is returned for any reason in response to a SleepQuery, make a determination here, return REJECT, and exit this function.
        // 

        // Do not enter sleep during file system processing
        if ( TryLockForSleep() )
        {
            // Perform the Sleep Mode pre-processing here as required
            if ( s_PrepareSleepCallback )
            {
                s_PrepareSleepCallback();
            }

            // Clears AwakeEvent immediately before permitting Sleep Mode
            s_AwakeEvent.ClearSignal();

            // Permits Sleep Mode, and stops the main thread immediately
            nn::applet::ReplySleepQuery(nn::applet::REPLY_ACCEPT);
            s_AwakeEvent.Wait();

            // Perform the processing when recovering from Sleep Mode here as required
            if ( s_AfterSleepCallback )
            {
                s_AfterSleepCallback();
            }

            UnlockForSleep();
        }
    }

    // Check termination request
    if ( nn::applet::IsExpectedToCloseApplication() )
    {
        s_IsExitRequired = true;
        return;
    }

    // Performs application transition processing

    // HOME Button processing (the graphics library must already be initialized)
    if ( nn::applet::IsExpectedToProcessHomeButton() )
    {
        // Here, if you recognize something like a flag and are to exit the block, you must delete the HOME button flag by calling nn::applet::ClearHomeButtonState, which is able to execute during the time that HOME button operation is prohibited. 
        // 

        if ( TryLock() )
        {
            if ( s_PrepareHomeButtonCallback )
            {
                s_PrepareHomeButtonCallback();
            }

            nn::applet::ProcessHomeButtonAndWait();

            Unlock();

            // If exiting from the HOME menu or if it is determined to end the application for a cause such as a low battery, exit from the application main loop and go to termination processing.
            // 
            if ( nn::applet::IsExpectedToCloseApplication() )
            {
                // No render rights are passed
                s_IsExitRequired = true;

                if ( s_AfterHomeButtonCallback )
                {
                    s_AfterHomeButtonCallback();
                }
                return;
            }

            if ( s_AfterHomeButtonCallback )
            {
                s_AfterHomeButtonCallback();
            }

            // Restore graphics settings
            RestoreGraphicSetting();
        }
    }

    // POWER Button processing (the graphics library must already be initialized)
    if ( nn::applet::IsExpectedToProcessPowerButton() )
    {
        // Promptly handle POWER button processing. All processing that is required before exiting an application (such as save data processing) should come after the control returns from nn::applet::ProcessPowerButtonAndWait().
        // 
        // 
        if ( s_PreparePowerButtonCallback )
        {
            s_PreparePowerButtonCallback();
        }

        nn::applet::ProcessPowerButtonAndWait();

        // If it is determined to end the application, exit from the application main loop and go to termination processing
        if ( nn::applet::IsExpectedToCloseApplication() )
        {
            // No render rights are passed
            s_IsExitRequired = true;
            return;
        }

        // Restore graphics settings
        RestoreGraphicSetting();
    }
}

bool TransitionHandler::IsExitRequired()
{
    return s_IsExitRequired;
}

void TransitionHandler::Lock()
{
    s_CriticalSection.Enter();
}

bool TransitionHandler::TryLock()
{
    return s_CriticalSection.TryEnter();
}

void TransitionHandler::Unlock()
{
    s_CriticalSection.Leave();
}

void TransitionHandler::LockForSleep()
{
    s_CriticalSectionForSleep.Enter();
}

bool TransitionHandler::TryLockForSleep()
{
    return s_CriticalSectionForSleep.TryEnter();
}

void TransitionHandler::UnlockForSleep()
{
    s_CriticalSectionForSleep.Leave();
}

/*------------------------------------------------------------------------*
    Callback called when there is a sleep query
 *------------------------------------------------------------------------*/
AppletQueryReply TransitionHandler::SleepQueryCallback( uptr arg )
{
    NN_UNUSED_VAR(arg);

    if ( !nn::applet::IsActive() )
    {
        // When Inactive, the main thread is in a stop state with applet::WaitForStarting(), and other threads should also be in a stop state (application-side implementation-dependent).
        // 
        return nn::applet::REPLY_ACCEPT;
    }
    else
    {
        // The main thread is suspended here because it will sleep after the necessary processing completes
        // 
        return nn::applet::REPLY_LATER;
    }
}

/*------------------------------------------------------------------------*
    Callback called when recovering from sleep
 *------------------------------------------------------------------------*/
void TransitionHandler::AwakeCallback( uptr arg )
{
    NN_UNUSED_VAR(arg);
    s_AwakeEvent.Signal();

    s_IsAfterWakeUp = true;
}
