# -*- coding: sjis -*-
require "optparse"
require "fileutils"
#===============================================================================
module NAR
#-------------------------------------------------------------------------------
# Marker bitmap class. Only 24-bit RGB for now.
class MarkerBmp
  attr_reader :file_name, :buf, :height, :width

  # Constructor.
  # file_name :: Specifies the name of the file to load.
  def initialize( file_name )
    @file_name = file_name
    load_image( file_name )
  end

  # Load a bitmap file.
  # file_name :: BMP filename
  def load_image( file_name )
    # Temporary buffer
    tmp_buf = nil

    # Load data from the file
    File.open( file_name, "rb") do |bmp|
      tmp_buf = bmp.read
    end

    # Size of the buffer.
    buf_size = tmp_buf[2, 4].unpack("l*").first

    # Get the width and height
    @width  = tmp_buf[18, 4].unpack("l*").first
    @height = tmp_buf[22, 4].unpack("l*").first

    # Get the image buffer.
    @buf = tmp_buf[54, buf_size]
  end

  # Get the RGB value at the specified point
  # x :: Pixel width
  # y :: Pixel height
  def get_color( x, y )
    # TODO: Check the range of X and Y
    line_size = @width * 3 + (4 - (@width * 3) % 4) % 4
    p = (@height - 1 - y) * line_size + x * 3

    # Since Ruby 1.9, you need to call the version of 'ord' that is implemented in Ruby 1.8.7.
    # If 'ord' doesn't exist, assume that ruby 1.8.6 is being used and load the data unchanged.
    if( @buf[p].respond_to? :ord )
      b = @buf[p].ord
      g = @buf[p+1].ord
      r = @buf[p+2].ord
    else
      b = @buf[p]
      g = @buf[p+1]
      r = @buf[p+2]
    end

    return [r, g, b]
  end
  
  # Get the average color of the specified image location and its neighboring pixels.
  # x :: Horizontal position (in pixels)
  # y :: Vertical position (in pixels)
  def get_average_color( x, y )
    arr_dx = [0,0,-1,0,1]
    arr_dy = [0,-1,0,1,0]

    cnt = arr_dx.size

    r = g = b = 0
    0.upto(cnt-1) do |point|

      cx = x + arr_dx[point]
      next if ( cx >= @width )

      cy = y + arr_dy[point]
      next if ( cy >= @height ) 
      
      c = get_color(cx,cy)
      r += c[0]
      g += c[1]
      b += c[2]
    end

    # Return the average color as an array.
    [r,g,b].map{|v| MarkerUtil.to_rgb5(v)/cnt}
    
  end
  
  # Create a table for adjusting the contrast.
  # v :: Contrast values between -127 and +127.
  def contrast_table(v)
    cont_tbl = []
    
    k = 1.0
    if v>0
      k=255.0/(255.0-2.0*v.to_f)
    else
      k=(255.0+2.0*v.to_f)/255.0
    end
    
    # Create a 0-255 table
    0.upto(255) do |i|
      v = k * (i.to_f-127.5) + 127.5
      cont_tbl << [0,[255, v].min].max.to_i
    end
    
    cont_tbl
  end

  # Adjust the contrast
  # image :: BMP image
  # v :: Contrast values between -127 and +127.
  def contrast!(v)
    ctbl = contrast_table(v)

    0.upto(@buf.size-1) do |i|
      @buf[i] = ctbl[@buf[i]]
    end

    self
  end

end

#-------------------------------------------------------------------------------
# Marker template class
class MarkerTemplate
  attr_reader :id
  attr_reader :aspect
  attr_reader :side_length
  attr_reader :pattern
  attr_reader :offset
  attr_reader :weight
  attr_reader :divide
  attr_reader :average
  attr_reader :stdevp
  attr_reader :bmp_file

  # id :: Marker ID
  # image :: Marker image
  # side_length :: Length of a side
  # corp :: Position information when a specified area is considered to be a marker
  #         This is specified as a hash with :left, :top, :width, and :height as keys
  # disable_offset :: Should an offset be used?
  # divide :: Number of divisions in the sampling grid
  def initialize( id, image, side_lenght=4.0, crop={}, disable_offset=false, divide=15 )

    # Filename
    @bmp_file = image.file_name

    # Set the ID
    @id = id

    # Set the divisions
    @divide = divide

    # Initialize crop if it is nil
    crop = {:left=>0, :top=>0, :width=>image.width, :height=>image.height} if !crop

    # If a width and height have not been specified, use the width and height of the bitmap unchanged
    l = crop.has_key?(:left)   ? crop[:left]   : 0
    t = crop.has_key?(:top)    ? crop[:top]    : 0
    w = crop.has_key?(:width)  ? crop[:width]  : image.width
    h = crop.has_key?(:height) ? crop[:height] : image.height

    # Calculate the aspect ratio
    @aspect = w*1.0/h

    # Length of a side 
    @side_length = side_lenght

    # Get the raw pattern that was sampled
    raw_pattern = get_raw_pattern(image, l, t, w, h )

    # Get the offset value
    if !disable_offset
      offset = raw_pattern.inject(0){|t,v| t + v} / raw_pattern.size
    else
      offset = 0
    end

    # Subtract the average from the pattern
    @pattern = raw_pattern.map{|v| v-offset}

    # Sum the pattern values
    sum_val  = @pattern.inject(0){|t,v| t + v}

    # Sum the squared pattern values
    sum_pow2 = @pattern.inject(0){|t,v| t + (v*v) }

    # Average the pattern values
    @average = sum_val*1.0 / @pattern.size

    # Standard deviation
    @stdevp  = Math.sqrt(sum_pow2 / @pattern.size - @average ** 2)

    # Set the average as the offset
    @offset = offset

    # Initialize the weight to 16
    @weight = []
    (@pattern.size/3).times{ @weight << 16 }
  end

  # Get a pattern from an image.
  # This returns an array with the points in the specified grid and the average of the four points around them.
  # image :: The target image
  # left   :: The left side of the marker
  # top    :: The top of the marker
  # width  :: The width of the marker
  # height :: The height of the marker
  def get_raw_pattern(image, left, top, width, height)
    raw_pattern = []

    # Simply create the grid point with the number of divisions and get the color, because we assume bitmaps are flat images.
    # 
    h_pitch = width  * 1.0 /(@divide + 1)
    v_pitch = height * 1.0 /(@divide + 1)

    1.upto(@divide) do |v|
      1.upto(@divide) do |h|
        # Get the average colors and turn them into a pattern.
        raw_pattern << image.get_average_color(left+(h*h_pitch).round, top+(v*v_pitch).round)
      end
    end
    # Return the pattern.
    raw_pattern.flatten
  end

  def get_divide_code(arr, tab_cnt)
    pattern_text = []
    1.upto(@divide) do |v|
      line = []
      1.upto(@divide) do |h|
        0.upto(2) do |i|
          line << arr[v*@divide+h+i]
        end
      end
      pattern_text << line.join(",")
    end

   tab = ""
   1.upto(tab_cnt){tab += "\t"}
    pattern_text.join("\n#{tab}")
  end
end

#-------------------------------------------------------------------------------
# Utility for exporting marker information
module MarkerUtil
  class << self
  private
    # Get the code for patterns.
    # template :: Template
    # tabs :: Tabs to indent
    def get_pattern_code( template, tabs = "" )
      code = ""
      template.pattern.each_with_index do |v,i|
        if i%(template.divide*3) == 0
          code << "\n#{tabs}"
        end
        code << sprintf("%#2d,", v)
      end
      code
    end
    
    
    # Get the code for weights.
    # template :: Template
    # tabs :: Tabs to indent
    def get_weight_code( template, tabs = "" )
      code = ""
      template.weight.each_with_index do |v,i|
        if i%(template.divide) == 0
          code << "\n#{tabs}"
        end
        code << sprintf("%#2d,", v)
      end
      code
    end
    
    # Get the code for multiple templates as an array.
    # templates :: Array of templates
    def get_templates_code( templates )
code=<<-_CODE_
/**
 * [Caution]
 * This code is a sample easily-generated template. (not for the product)
 * Use marker_maker demo and replace this template code.
 */
mw::nar::MarkerTemplate_st ga_Markers[#{templates.size}] ={
#{
templates.inject("") do |r,t|
  "#{r}#{get_template_code_part(t).chomp},\n"
end
}
_CODE_
code + "}; // ga_Markers[#{templates.size}]"
    end
    
    # Get the code for a single template.
    # template :: Template
    def get_single_template_code( template )
code=<<-_CODE_
/**
 * [Caution]
 * This code is a sample easily-generated template. (not for the product)
 * Use marker_maker demo and replace this template code.
 */
mw::nar::MarkerTemplate_st g_Marker =
#{get_template_code_part(template)}
_CODE_
code.strip + ";"
    end
    
    # Get the code for part of a template.
    # template :: Template
    def get_template_code_part( template )
code=<<-_CODE_
{ // #{File.basename(template.bmp_file.gsub(/\.tmp/,""))}
#{template.id}, 		// id
#{template.aspect}, // aspect
#{template.side_length},	// Length of a side
{
#{get_pattern_code(template, "\t")}
	#{template.offset},
	#{template.average},
	#{template.stdevp},
},
}
_CODE_
    end

  public
    # Export a table for review.
    # template :: Template
    def get_pattern_table( template )
      code = ""
      vals = []
      template.pattern.each_with_index do |v,i|
        vals << v
        if vals.size == 3
          puts vals.join("\t")
          vals.clear
        end
      end
      code
    end

    # Export data in a table for review.
    # template :: Template
    # ofile :: Output file
    def print_pattern_table( template, ofile=nil )
      o = ofile || $stdout
      o.puts File.basename(template.bmp_file)
      o.puts get_pattern_table(template)
      o.puts ""
    end

    # Export the code for a single template.
    # template :: Template
    # ofile :: Output file
    def print_single_template_code( template, ofile=nil)
      o = ofile || $stdout
      o.puts get_single_template_code(template)
    end

    # Export the code for multiple templates.
    # templates :: Array of templates
    # ofile :: Output file
    def print_templates_code(templates, ofile=nil)
      o = ofile || $stdout
      o.puts get_templates_code(templates)
    end
    

    # Save a template as a BMP file.
    # template :: Template
    # out_dir :: Output folder
    def save_template_as_bmp(template, out_dir=nil)
      
      fname = File.join(File.dirname(template.bmp_file), "__#{File.basename(template.bmp_file)}")
      fname = File.join(out_dir,fname) if out_dir
 
      # Width and height
      width = height = template.divide
      
      # Add and return the averages
      field = []
      col = []
      template.pattern.each_with_index do |c,cnt|
        c+=template.offset
        idx = ( cnt / 3 ) % width
        if cnt % 3 == 0
          col<<0
          col[ idx ] |= c << 10
        elsif cnt % 3 == 1
          col[ idx ] |= c << 5
        elsif cnt % 3 == 2
          col[ idx ] |= c
          if idx == width - 1
            rest = 4 - col.size % 4
            rest.times do
              col.push(0)
            end
            field = col + field
            col = []
          end
        end
      end

      File.open(fname, "wb") do |file|
        file.write(["BM", 54 + width * height * 2, 54].pack("a2Vx4V"))
        file.write([40, width, height, 1, 16].pack("VVVvvx24"))
        file.write(field.pack("v*"))
      end
      
    end
    
    # Convert an 8-bit color into a 5-bit color
    # val :: Value to convert
    def to_rgb5( val )
      val >> 3
    end
    
    # Revert a 5-bit color into an 8-bit color
    # val :: Value to convert
    def to_rgb8( val )
      val << 3
    end
    
  end # class<self
  
end
end # NAR

#===============================================================================
# Generate a template from a BMP image.
# id :: Marker ID
# bmp :: Bitmap
# side_length :: Length of a side
# crop :: Specifies where to crop the image (in pixels)
#         {:left=>Left, :top=>Top, :width=>Width, :height=>Height}
# contrast_value :: Contrast
# disable_offset :: Enable/disable the offset
def create_template(id, bmp, side_lenght, crop, contrast_value, disable_offset)
  # Load the BMP
  marker_bmp = NAR::MarkerBmp.new(bmp)

  # Adjust the contrast
  if contrast_value != 0
    marker_bmp.contrast!(contrast_value)
  end

  # Generate a template from the BMP image
  template = NAR::MarkerTemplate.new( id, marker_bmp, side_lenght, crop, disable_offset )
end

# Export multiple templates.
# templates :: Array of templates
# ofile :: Output file
def print_templates( templates, ofile )
  NAR::MarkerUtil.print_templates_code( templates, ofile )
end

# Export a template.
# template :: Template
# ofile :: Output file
def print_single_template( template, ofile, is_save_bmp = false )
  NAR::MarkerUtil.print_single_template_code( template, ofile )
end

# Template operations
# img_file :: Path to an input image
def proc_template_img( img_file )
    # If it is a PNG:
    if /\.png$/ =~ img_file
      # If the image is a PNG, convert it into a temporary BMP
      tmp_file = img_file + ".tmp"
      system( "png2bmp.exe #{img_file} #{tmp_file}" )
      # Pass the converted BMP file to a block
     yield tmp_file

      # Delete the temporary BMP file
      FileUtils.rm_rf( tmp_file )

    # If it is a BMP:
    else
      
      # Pass the converted BMP file to a block
      yield img_file
    end

end

#==============================================================================
# Main
# Usage:
# ruby ar_marker_maker.rb -i path/to/template.bmp
# 
# If there are margins:
# ruby ar_marker_maker.rb "%f" -i sample_marker/hatena.png  --output=hatena.txt --crop=47,71,543,869
# Option details:
# displayed in ruby ar_marker_maker.rb --help
#==============================================================================
params = {:id=>0, :side_length=>4.0, :contrast=>0, :save_bmp=>false, :disable_offset=>false}

# Parse arguments
ARGV.options do |opt|
  opt.on('-v','--version','Show the version number and quit') do |val|
    puts("NARMarkerTemplateMaker 1.1.0")
  end
  
  opt.on('-i','--input=path/to/imgfile or imgdir', 'Template image file or directory.') do |val|
    params.merge!({:input => val})
  end

  opt.on('-o','--output=path/to/outputfile', 'Output template code file.') do |val|
    params.merge!({:output => val})
  end

  opt.on('--id=id','Template id.') do |val|
    params.merge!({:id => val})
  end

  opt.on('--side_length=val','Side length value.') do |val|
    params.merge!({:side_length => val})
  end

  opt.on('--crop=l,t,w,h', 'Crop image option. l:left t:top w:width h:height') do |val|
    c = val.scan(/\d+/)
    raise "Invalid crop argument." if( c.size != 4 )
    c.map!{|v| v.to_i}
    params.merge!({:crop=> {:left=>c[0], :top=>c[1], :width=>c[2], :height=>c[3]}})
  end

  opt.on('--contrast=value', 'Adjust contrast value 0:normal(default) -100:minimum 100:maxmum') do |val|
    params.merge!({:contrast => val.to_i})
  end

  opt.on('--save_bmp', 'Save created pattern as bmp file') do |val|
    params.merge!({:save_bmp => val})
  end

  opt.on('--disable_offset', 'Disable offset of pattern') do |val|
    params.merge!({:disable_offset => val})
  end

  opt.parse!
end

if !params[:input]
  puts "HELP: ruby ar_marker_maker.rb --help"
end


# If an output file is specified, open it
ofile = nil
if params[:output]
  ofile = File.open(params[:output],"w")
end

begin
  # Create templates from a specified file or the BMP files in a specified folder
  templates = []

  if params[:input] && File.file?(params[:input]) # For files:

    img_file = params[:input]
    template = nil

    # Create a template from an image file
    proc_template_img( img_file ) do |bmp_file|
      template = create_template(params[:id], bmp_file, params[:side_length], params[:crop], params[:contrast], params[:disable_offset])
    end

    print_single_template( template, ofile )
    if( params[:save_bmp] )
      NAR::MarkerUtil.save_template_as_bmp( template )
    end

  elsif params[:input] && File.directory?(params[:input]) # For folders:
    id = params[:id].to_i
    # Search for BMP and PNG files in the directory
    Dir.glob(File.join(params[:input], "*.{bmp,png}")) do |img_file|
      next if /^__/ =~ File.basename(img_file)

      # Create a template from an image file
      proc_template_img( img_file ) do |bmp_file|
        templates << create_template(id, bmp_file, params[:side_length], params[:crop], params[:contrast], params[:disable_offset])
      end

      id+=1
    end
    # Export templates
    print_templates( templates, ofile )
    if( params[:save_bmp] )
      templates.each do |t|
        NAR::MarkerUtil.save_template_as_bmp( t )
      end
    end
  end
#rescue => e
#  puts "Failed to create templates!"
#  puts e
#ensure
  ofile.close if ofile && !ofile.closed?
end

